import base64
import copy
import os
import threading
import uuid
from io import BytesIO

from flask import jsonify, send_from_directory
from werkzeug.utils import secure_filename

from fedn.common.config import get_controller_config, get_network_config
from fedn.network.combiner.interfaces import (CombinerInterface,
                                              CombinerUnavailableError)
from fedn.network.state import ReducerState, ReducerStateToString
from fedn.utils.checksum import sha
from fedn.utils.plots import Plot

__all__ = ("API",)


class API:
    """The API class is a wrapper for the statestore. It is used to expose the statestore to the network API."""

    def __init__(self, statestore, control):
        self.statestore = statestore
        self.control = control
        self.name = "api"

    def _to_dict(self):
        """Convert the object to a dict.

        ::return: The object as a dict.
        ::rtype: dict
        """
        data = {"name": self.name}
        return data

    def _allowed_file_extension(
        self, filename, ALLOWED_EXTENSIONS={"gz", "bz2", "tar", "zip", "tgz"}
    ):
        """Check if file extension is allowed.

        :param filename: The filename to check.
        :type filename: str
        :return: True and extension str if file extension is allowed, else False and None.
        :rtype: Tuple (bool, str)
        """
        if "." in filename:
            extension = filename.rsplit(".", 1)[1].lower()
            if extension in ALLOWED_EXTENSIONS:
                return (True, extension)

        return (False, None)

    def get_clients(self, limit=None, skip=None, status=False):
        """Get all clients from the statestore.

        :return: All clients as a json response.
        :rtype: :class:`flask.Response`
        """
        # Will return list of ObjectId
        response = self.statestore.list_clients(limit, skip, status)

        arr = []

        for element in response["result"]:
            obj = {
                "id": element["name"],
                "combiner": element["combiner"],
                "combiner_preferred": element["combiner_preferred"],
                "ip": element["ip"],
                "status": element["status"],
                "last_seen": element["last_seen"] if "last_seen" in element else "",
            }

            arr.append(obj)

        result = {"result": arr, "count": response["count"]}

        return jsonify(result)

    def get_all_combiners(self, limit=None, skip=None):
        """Get all combiners from the statestore.

        :return: All combiners as a json response.
        :rtype: :class:`flask.Response`
        """
        # Will return list of ObjectId
        projection = {"name": True, "updated_at": True}
        response = self.statestore.get_combiners(limit, skip, projection=projection)
        arr = []
        for element in response["result"]:
            obj = {
                "name": element["name"],
                "updated_at": element["updated_at"],
            }

            arr.append(obj)

        result = {"result": arr, "count": response["count"]}

        return jsonify(result)

    def get_combiner(self, combiner_id):
        """Get a combiner from the statestore.

        :param combiner_id: The combiner id to get.
        :type combiner_id: str
        :return: The combiner info dict as a json response.
        :rtype: :class:`flask.Response`
        """
        # Will return ObjectId
        object = self.statestore.get_combiner(combiner_id)
        payload = {}
        id = object["name"]
        info = {
            "address": object["address"],
            "fqdn": object["fqdn"],
            "parent_reducer": object["parent"]["name"],
            "port": object["port"],
            "updated_at": object["updated_at"],
        }
        payload[id] = info

        return jsonify(payload)

    def get_all_sessions(self, limit=None, skip=None):
        """Get all sessions from the statestore.

        :return: All sessions as a json response.
        :rtype: :class:`flask.Response`
        """
        sessions_object = self.statestore.get_sessions(limit, skip)
        if sessions_object is None:
            return (
                jsonify({"success": False, "message": "No sessions found."}),
                404,
            )
        arr = []
        for element in sessions_object["result"]:
            obj = element["session_config"][0]
            arr.append(obj)

        result = {"result": arr, "count": sessions_object["count"]}

        return jsonify(result)

    def get_session(self, session_id):
        """Get a session from the statestore.

        :param session_id: The session id to get.
        :type session_id: str
        :return: The session info dict as a json response.
        :rtype: :class:`flask.Response`
        """
        session_object = self.statestore.get_session(session_id)
        if session_object is None:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": f"Session {session_id} not found.",
                    }
                ),
                404,
            )
        payload = {}
        id = session_object["session_id"]
        info = session_object["session_config"][0]
        status = session_object["status"]
        payload[id] = info
        payload['status'] = status
        return jsonify(payload)

    def set_active_compute_package(self, id: str):

        success = self.statestore.set_active_compute_package(id)

        if not success:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "Failed to set compute package.",
                    }
                ),
                400,
            )

        return jsonify({"success": True, "message": "Compute package set."})

    def set_compute_package(self, file, helper_type: str, name: str = None, description: str = None):
        """Set the compute package in the statestore.

        :param file: The compute package to set.
        :type file: file
        :return: A json response with success or failure message.
        :rtype: :class:`flask.Response`
        """

        if (
            self.control.state() == ReducerState.instructing
            or self.control.state() == ReducerState.monitoring
        ):
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "Reducer is in instructing or monitoring state."
                        "Cannot set compute package.",
                    }
                ),
                400,
            )

        if file is None:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "No file provided.",
                    }
                ),
                404,
            )

        success, extension = self._allowed_file_extension(file.filename)

        if not success:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": f"File extension {extension} not allowed.",
                    }
                ),
                404,
            )

        file_name = file.filename
        storage_file_name = secure_filename(f"{str(uuid.uuid4())}.{extension}")

        file_path = os.path.join("/app/client/package/", storage_file_name)
        file.save(file_path)

        self.control.set_compute_package(storage_file_name, file_path)
        success = self.statestore.set_compute_package(file_name, storage_file_name, helper_type, name, description)

        if not success:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "Failed to set compute package.",
                    }
                ),
                400,
            )

        return jsonify({"success": True, "message": "Compute package set."})

    def _get_compute_package_name(self):
        """Get the compute package name from the statestore.

        :return: The compute package name.
        :rtype: str
        """
        package_objects = self.statestore.get_compute_package()
        if package_objects is None:
            message = "No compute package found."
            return None, message
        else:
            try:
                name = package_objects["storage_file_name"]
            except KeyError as e:
                message = "No compute package found. Key error."
                print(e)
                return None, message
            return name, "success"

    def get_compute_package(self):
        """Get the compute package from the statestore.

        :return: The compute package as a json response.
        :rtype: :class:`flask.Response`
        """
        result = self.statestore.get_compute_package()
        if result is None:
            return (
                jsonify(
                    {"success": False, "message": "No compute package found."}
                ),
                404,
            )

        obj = {
            "id": result["id"] if "id" in result else "",
            "file_name": result["file_name"],
            "helper": result["helper"],
            "committed_at": result["committed_at"],
            "storage_file_name": result["storage_file_name"] if "storage_file_name" in result else "",
            "name": result["name"] if "name" in result else "",
            "description": result["description"] if "description" in result else "",
        }

        return jsonify(obj)

    def list_compute_packages(self, limit: str = None, skip: str = None, include_active: str = None):
        """Get paginated list of compute packages from the statestore.
        :param limit: The number of compute packages to return.
        :type limit: str
        :param skip: The number of compute packages to skip.
        :type skip: str
        :param include_active: Whether to include the active compute package or not.
        :type include_active: str
        :return: All compute packages as a json response.
        :rtype: :class:`flask.Response`
        """

        if limit is not None and skip is not None:
            limit = int(limit)
            skip = int(skip)

        include_active: bool = include_active == "true"

        result = self.statestore.list_compute_packages(limit, skip)
        if result is None:
            return (
                jsonify(
                    {"success": False, "message": "No compute packages found."}
                ),
                404,
            )

        active_package_id: str = None

        if include_active:
            active_package = self.statestore.get_compute_package()

            if active_package is not None:
                active_package_id = active_package["id"] if "id" in active_package else ""

        if include_active:
            arr = [
                {
                    "id": element["id"] if "id" in element else "",
                    "file_name": element["file_name"],
                    "helper": element["helper"],
                    "committed_at": element["committed_at"],
                    "storage_file_name": element["storage_file_name"] if "storage_file_name" in element else "",
                    "name": element["name"] if "name" in element else "",
                    "description": element["description"] if "description" in element else "",
                    "active": "id" in element and element["id"] == active_package_id,
                }
                for element in result["result"]
            ]
        else:
            arr = [
                {
                    "id": element["id"] if "id" in element else "",
                    "file_name": element["file_name"],
                    "helper": element["helper"],
                    "committed_at": element["committed_at"],
                    "storage_file_name": element["storage_file_name"] if "storage_file_name" in element else "",
                    "name": element["name"] if "name" in element else "",
                    "description": element["description"] if "description" in element else "",
                }
                for element in result["result"]
            ]

        result = {"result": arr, "count": result["count"]}
        return jsonify(result)

    def download_compute_package(self, name):
        """Download the compute package.

        :return: The compute package as a json object.
        :rtype: :class:`flask.Response`
        """
        if name is None:
            name, message = self._get_compute_package_name()
            if name is None:
                return jsonify({"success": False, "message": message}), 404
        try:
            mutex = threading.Lock()
            mutex.acquire()
            # TODO: make configurable, perhaps in config.py or package.py
            return send_from_directory(
                "/app/client/package/", name, as_attachment=True
            )
        except Exception:
            try:
                data = self.control.get_compute_package(name)
                # TODO: make configurable, perhaps in config.py or package.py
                file_path = os.path.join("/app/client/package/", name)
                with open(file_path, "wb") as fh:
                    fh.write(data)
                # TODO: make configurable, perhaps in config.py or package.py
                return send_from_directory(
                    "/app/client/package/", name, as_attachment=True
                )
            except Exception:
                raise
        finally:
            mutex.release()

    def _create_checksum(self, name=None):
        """Create the checksum of the compute package.

        :param name: The name of the compute package.
        :type name: str
        :return: Success or failure boolean, message and the checksum.
        :rtype: bool, str, str
        """

        if name is None:
            name, message = self._get_compute_package_name()
            if name is None:
                return False, message, ""
        file_path = os.path.join(
            "/app/client/package/", name
        )  # TODO: make configurable, perhaps in config.py or package.py
        try:
            sum = str(sha(file_path))
        except FileNotFoundError:
            sum = ""
            message = "File not found."
        return True, message, sum

    def get_checksum(self, name):
        """Get the checksum of the compute package.

        :param name: The name of the compute package.
        :type name: str
        :return: The checksum as a json object.
        :rtype: :py:class:`flask.Response`
        """

        success, message, sum = self._create_checksum(name)
        if not success:
            return jsonify({"success": False, "message": message}), 404
        payload = {"checksum": sum}

        return jsonify(payload)

    def get_controller_status(self):
        """Get the status of the controller.

        :return: The status of the controller as a json object.
        :rtype: :py:class:`flask.Response`
        """
        return jsonify({"state": ReducerStateToString(self.control.state())})

    def get_events(self, **kwargs):
        """Get the events of the federated network.

        :return: The events as a json object.
        :rtype: :py:class:`flask.Response`
        """
        response = self.statestore.get_events(**kwargs)

        result = response["result"]
        if result is None:
            return (
                jsonify({"success": False, "message": "No events found."}),
                404,
            )

        events = []
        for evt in result:
            events.append(evt)

        return jsonify({"result": events, "count": response["count"]})

    def get_all_validations(self, **kwargs):
        """Get all validations from the statestore.

        :return: All validations as a json response.
        :rtype: :class:`flask.Response`
        """
        validations_objects = self.statestore.get_validations(**kwargs)
        if validations_objects is None:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "No validations found.",
                        "filter_used": kwargs,
                    }
                ),
                404,
            )
        payload = {}
        for object in validations_objects:
            id = str(object["_id"])
            info = {
                "model_id": object["modelId"],
                "data": object["data"],
                "timestamp": object["timestamp"],
                "meta": object["meta"],
                "sender": object["sender"],
                "receiver": object["receiver"],
            }
            payload[id] = info
        return jsonify(payload)

    def add_combiner(
        self, combiner_id, secure_grpc, address, remote_addr, fqdn, port
    ):
        """Add a combiner to the network.

        :param combiner_id: The combiner id to add.
        :type combiner_id: str
        :param secure_grpc: Whether to use secure grpc or not.
        :type secure_grpc: bool
        :param name: The name of the combiner.
        :type name: str
        :param address: The address of the combiner.
        :type address: str
        :param remote_addr: The remote address of the combiner.
        :type remote_addr: str
        :param fqdn: The fqdn of the combiner.
        :type fqdn: str
        :param port: The port of the combiner.
        :type port: int
        :return: Config of the combiner as a json response.
        :rtype: :class:`flask.Response`
        """
        # TODO: Any more required check for config? Formerly based on status: "retry"
        if not self.control.idle():
            return jsonify(
                {
                    "success": False,
                    "status": "retry",
                    "message": "Conroller is not in idle state, try again later. ",
                }
            )
        # Check if combiner already exists
        combiner = self.control.network.get_combiner(combiner_id)
        if not combiner:
            if secure_grpc == "True":
                certificate, key = self.certificate_manager.get_or_create(
                    address
                ).get_keypair_raw()
                _ = base64.b64encode(certificate)
                _ = base64.b64encode(key)

            else:
                certificate = None
                key = None

            combiner_interface = CombinerInterface(
                parent=self._to_dict(),
                name=combiner_id,
                address=address,
                fqdn=fqdn,
                port=port,
                certificate=copy.deepcopy(certificate),
                key=copy.deepcopy(key),
                ip=remote_addr,
            )

            self.control.network.add_combiner(combiner_interface)

        # Check combiner now exists
        combiner = self.control.network.get_combiner(combiner_id)
        if not combiner:
            return jsonify(
                {"success": False, "message": "Combiner not added."}
            )

        payload = {
            "success": True,
            "message": "Combiner added successfully.",
            "status": "added",
            "storage": self.statestore.get_storage_backend(),
            "statestore": self.statestore.get_config(),
            "certificate": combiner.get_certificate(),
            "key": combiner.get_key(),
        }

        return jsonify(payload)

    def add_client(self, client_id, preferred_combiner, remote_addr):
        """Add a client to the network.

        :param client_id: The client id to add.
        :type client_id: str
        :param preferred_combiner: The preferred combiner for the client.If None, the combiner will be chosen based on availability.
        :type preferred_combiner: str
        :return: A json response with combiner assignment config.
        :rtype: :class:`flask.Response`
        """
        # Check if package has been set
        package_object = self.statestore.get_compute_package()
        if package_object is None:
            return (
                jsonify(
                    {
                        "success": False,
                        "status": "retry",
                        "message": "No compute package found. Set package in controller.",
                    }
                ),
                203,
            )

        # Assign client to combiner
        if preferred_combiner:
            combiner = self.control.network.get_combiner(preferred_combiner)
            if combiner is None:
                return (
                    jsonify(
                        {
                            "success": False,
                            "message": f"Combiner {preferred_combiner} not found or unavailable.",
                        }
                    ),
                    400,
                )
        else:
            combiner = self.control.network.find_available_combiner()
            if combiner is None:
                return (
                    jsonify(
                        {"success": False, "message": "No combiner available."}
                    ),
                    400,
                )

        client_config = {
            "name": client_id,
            "combiner_preferred": preferred_combiner,
            "combiner": combiner.name,
            "ip": remote_addr,
            "status": "available",
        }
        # Add client to network
        self.control.network.add_client(client_config)

        # Setup response containing information about the combiner for assinging the client
        if combiner.certificate:
            cert_b64 = base64.b64encode(combiner.certificate)
            cert = str(cert_b64).split("'")[1]
        else:
            cert = None

        payload = {
            "status": "assigned",
            "host": combiner.address,
            "fqdn": combiner.fqdn,
            "package": "remote",  # TODO: Make this configurable
            "ip": combiner.ip,
            "port": combiner.port,
            "certificate": cert,
            "helper_type": self.control.statestore.get_helper(),
        }
        print("Seding payload: ", payload, flush=True)

        return jsonify(payload)

    def get_initial_model(self):
        """Get the initial model from the statestore.

        :return: The initial model as a json response.
        :rtype: :class:`flask.Response`
        """
        model_id = self.statestore.get_initial_model()
        payload = {"model_id": model_id}
        return jsonify(payload)

    def set_initial_model(self, file):
        """Add an initial model to the network.

        :param file: The initial model to add.
        :type file: file
        :return: A json response with success or failure message.
        :rtype: :class:`flask.Response`
        """
        try:
            object = BytesIO()
            object.seek(0, 0)
            file.seek(0)
            object.write(file.read())
            helper = self.control.get_helper()
            object.seek(0)
            model = helper.load(object)
            self.control.commit(file.filename, model)
        except Exception as e:
            print(e, flush=True)
            return jsonify({"success": False, "message": e})

        return jsonify(
            {"success": True, "message": "Initial model added successfully."}
        )

    def get_latest_model(self):
        """Get the latest model from the statestore.

        :return: The initial model as a json response.
        :rtype: :class:`flask.Response`
        """
        if self.statestore.get_latest_model():
            model_id = self.statestore.get_latest_model()
            payload = {"model_id": model_id}
            return jsonify(payload)
        else:
            return jsonify(
                {"success": False, "message": "No initial model set."}
            )

    def set_current_model(self, model_id: str):
        """Set the active model in the statestore.

        :param model_id: The model id to set.
        :type model_id: str
        :return: A json response with success or failure message.
        :rtype: :class:`flask.Response`
        """
        success = self.statestore.set_current_model(model_id)

        if not success:
            return (
                jsonify(
                    {
                        "success": False,
                        "message": "Failed to set active model.",
                    }
                ),
                400,
            )

        return jsonify({"success": True, "message": "Active model set."})

    def get_models(self, session_id: str = None, limit: str = None, skip: str = None, include_active: str = None):
        result = self.statestore.list_models(session_id, limit, skip)

        if result is None:
            return (
                jsonify({"success": False, "message": "No models found."}),
                404,
            )

        include_active: bool = include_active == "true"

        if include_active:

            latest_model = self.statestore.get_latest_model()

            arr = [
                {
                    "committed_at": element["committed_at"],
                    "model": element["model"],
                    "session_id": element["session_id"],
                    "active": element["model"] == latest_model,
                }
                for element in result["result"]
            ]
        else:
            arr = [
                {
                    "committed_at": element["committed_at"],
                    "model": element["model"],
                    "session_id": element["session_id"],
                }
                for element in result["result"]
            ]

        result = {"result": arr, "count": result["count"]}

        return jsonify(result)

    def get_model(self, model_id: str):
        result = self.statestore.get_model(model_id)

        if result is None:
            return (
                jsonify({"success": False, "message": "No model found."}),
                404,
            )

        payload = {
            "committed_at": result["committed_at"],
            "parent_model": result["parent_model"],
            "model": result["model"],
            "session_id": result["session_id"],
        }

        return jsonify(payload)

    def get_model_trail(self):
        """Get the model trail for a given session.

        :param session: The session id to get the model trail for.
        :type session: str
        :return: The model trail for the given session as a json response.
        :rtype: :class:`flask.Response`
        """
        model_info = self.statestore.get_model_trail()
        if model_info:
            return jsonify(model_info)
        else:
            return jsonify(
                {"success": False, "message": "No model trail available."}
            )

    def get_model_ancestors(self, model_id: str, limit: str = None):
        """Get the model ancestors for a given model.

        :param model_id: The model id to get the model ancestors for.
        :type model_id: str
        :param limit: The number of ancestors to return.
        :type limit: str
        :return: The model ancestors for the given model as a json response.
        :rtype: :class:`flask.Response`
        """
        if model_id is None:
            return jsonify(
                {"success": False, "message": "No model id provided."}
            )

        limit: int = int(limit) if limit is not None else 10  # if limit is None, default to 10

        response = self.statestore.get_model_ancestors(model_id, limit)
        if response:

            arr: list = []

            for element in response:
                obj = {
                    "model": element["model"],
                    "committed_at": element["committed_at"],
                    "session_id": element["session_id"],
                    "parent_model": element["parent_model"],
                }
                arr.append(obj)

            result = {"result": arr}

            return jsonify(result)
        else:
            return jsonify(
                {"success": False, "message": "No model ancestors available."}
            )

    def get_model_descendants(self, model_id: str, limit: str = None):
        """Get the model descendants for a given model.

        :param model_id: The model id to get the model descendants for.
        :type model_id: str
        :param limit: The number of descendants to return.
        :type limit: str
        :return: The model descendants for the given model as a json response.
        :rtype: :class:`flask.Response`
        """

        if model_id is None:
            return jsonify(
                {"success": False, "message": "No model id provided."}
            )

        limit: int = int(limit) if limit is not None else 10

        response: list = self.statestore.get_model_descendants(model_id, limit)

        if response:

            arr: list = []

            for element in response:
                obj = {
                    "model": element["model"],
                    "committed_at": element["committed_at"],
                    "session_id": element["session_id"],
                    "parent_model": element["parent_model"],
                }
                arr.append(obj)

            result = {"result": arr}

            return jsonify(result)
        else:
            return jsonify(
                {"success": False, "message": "No model descendants available."}
            )

    def get_all_rounds(self):
        """Get all rounds.

        :return: The rounds as json response.
        :rtype: :class:`flask.Response`
        """
        rounds_objects = self.statestore.get_rounds()
        if rounds_objects is None:
            jsonify({"success": False, "message": "No rounds available."})
        payload = {}
        for object in rounds_objects:
            id = object["round_id"]
            if "reducer" in object.keys():
                reducer = object["reducer"]
            else:
                reducer = None
            if "combiners" in object.keys():
                combiners = object["combiners"]
            else:
                combiners = None

            info = {
                "reducer": reducer,
                "combiners": combiners,
            }
            payload[id] = info
        else:
            return jsonify(payload)

    def get_round(self, round_id):
        """Get a round.

        :param round_id: The round id to get.
        :type round_id: str
        :return: The round as json response.
        :rtype: :class:`flask.Response`
        """
        round_object = self.statestore.get_round(round_id)
        if round_object is None:
            return jsonify({"success": False, "message": "Round not found."})
        payload = {
            'round_id': round_object['round_id'],
            'combiners': round_object['combiners'],
        }
        return jsonify(payload)

    def get_client_config(self, checksum=True):
        """Get the client config.

        :return: The client config as json response.
        :rtype: :py:class:`flask.Response`
        """
        config = get_controller_config()
        network_id = get_network_config()
        port = config["port"]
        host = config["host"]
        payload = {
            "network_id": network_id,
            "discover_host": host,
            "discover_port": port,
        }
        if checksum:
            success, _, checksum_str = self._create_checksum()
            if success:
                payload["checksum"] = checksum_str
        return jsonify(payload)

    def get_plot_data(self, feature=None):
        """Get plot data.

        :return: The plot data as json response.
        :rtype: :py:class:`flask.Response`
        """

        plot = Plot(self.control.statestore)

        try:
            valid_metrics = plot.fetch_valid_metrics()
            feature = feature or valid_metrics[0]
            box_plot = plot.create_box_plot(feature)
        except Exception as e:
            valid_metrics = None
            box_plot = None
            print(e, flush=True)

        result = {
            "valid_metrics": valid_metrics,
            "box_plot": box_plot,
        }

        return jsonify(result)

    def list_combiners_data(self, combiners):
        """Get combiners data.

        :param combiners: The combiners to get data for.
        :type combiners: list
        :return: The combiners data as json response.
        :rtype: :py:class:`flask.Response`
        """

        response = self.statestore.list_combiners_data(combiners)

        arr = []

        # order list by combiner name
        for element in response:

            obj = {
                "combiner": element["_id"],
                "count": element["count"],
            }

            arr.append(obj)

        result = {"result": arr}

        return jsonify(result)

    def start_session(
        self,
        session_id,
        aggregator='fedavg',
        model_id=None,
        rounds=5,
        round_timeout=180,
        round_buffer_size=-1,
        delete_models=True,
        validate=True,
        helper="numpyhelper",
        min_clients=1,
        requested_clients=8,
    ):
        """Start a session.

        :param session_id: The session id to start.
        :type session_id: str
        :param aggregator: The aggregator plugin to use.
        :type aggregator: str
        :param initial_model: The initial model for the session.
        :type initial_model: str
        :param rounds: The number of rounds to perform.
        :type rounds: int
        :param round_timeout: The round timeout to use in seconds.
        :type round_timeout: int
        :param round_buffer_size: The round buffer size to use.
        :type round_buffer_size: int
        :param delete_models: Whether to delete models after each round at combiner (save storage).
        :type delete_models: bool
        :param validate: Whether to validate the model after each round.
        :type validate: bool
        :param min_clients: The minimum number of clients required.
        :type min_clients: int
        :param requested_clients: The requested number of clients.
        :type requested_clients: int
        :return: A json response with success or failure message and session config.
        :rtype: :class:`flask.Response`
        """
        # Check if session already exists
        session = self.statestore.get_session(session_id)
        if session:
            return jsonify(
                {"success": False, "message": "Session already exists."}
            )

        # Check if session is running
        if self.control.state() == ReducerState.monitoring:
            return jsonify(
                {"success": False, "message": "A session is already running."}
            )

        # Check if compute package is set
        if not self.statestore.get_compute_package():
            return jsonify(
                {
                    "success": False,
                    "message": "No compute package set. Set compute package before starting session.",
                }
            )

        # Check that initial (seed) model is set
        if not self.statestore.get_initial_model():
            return jsonify(
                {
                    "success": False,
                    "message": "No initial model set. Set initial model before starting session.",
                }
            )

        # Check available clients per combiner
        clients_available = 0
        for combiner in self.control.network.get_combiners():
            try:
                nr_active_clients = len(combiner.list_active_clients())
                clients_available = clients_available + int(nr_active_clients)
            except CombinerUnavailableError as e:
                # TODO: Handle unavailable combiner, stop session or continue?
                print("COMBINER UNAVAILABLE: {}".format(e), flush=True)
                continue

        if clients_available < min_clients:
            return jsonify(
                {
                    "success": False,
                    "message": "Not enough clients available to start session.",
                }
            )

        # Check if validate is string and convert to bool
        if isinstance(validate, str):
            if validate.lower() == "true":
                validate = True
            else:
                validate = False

        # Get lastest model as initial model for session
        if not model_id:
            model_id = self.statestore.get_latest_model()

        # Setup session config
        session_config = {
            "session_id": session_id if session_id else str(uuid.uuid4()),
            "aggregator": aggregator,
            "round_timeout": round_timeout,
            "buffer_size": round_buffer_size,
            "model_id": model_id,
            "rounds": rounds,
            "delete_models_storage": delete_models,
            "clients_required": min_clients,
            "clients_requested": requested_clients,
            "task": (""),
            "validate": validate,
            "helper_type": helper,
        }

        # Start session
        threading.Thread(
            target=self.control.session, args=(session_config,)
        ).start()

        # Return success response
        return jsonify(
            {
                "success": True,
                "message": "Session started successfully.",
                "config": session_config,
            }
        )



import json
from datetime import datetime

import numpy
import plotly
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from fedn.network.storage.statestore.mongostatestore import MongoStateStore


class Plot:
    """

    """

    def __init__(self, statestore):
        try:
            statestore_config = statestore.get_config()
            statestore = MongoStateStore(
                statestore_config['network_id'], statestore_config['mongo_config'])
            self.mdb = statestore.connect()
            self.status = self.mdb['control.status']
            self.round_time = self.mdb["control.round_time"]
            self.combiner_round_time = self.mdb["control.combiner_round_time"]
            self.psutil_usage = self.mdb["control.psutil_monitoring"]
            self.network_clients = self.mdb["network.clients"]

        except Exception as e:
            print("FAILED TO CONNECT TO MONGO, {}".format(e), flush=True)
            self.collection = None
            raise

    # plot metrics from DB
    def _scalar_metrics(self, metrics):
        """ Extract all scalar valued metrics from a MODEL_VALIDATON. """

        data = json.loads(metrics['data'])
        data = json.loads(data['data'])

        valid_metrics = []
        for metric, val in data.items():
            # If it can be converted to a float it is a valid, scalar metric
            try:
                val = float(val)
                valid_metrics.append(metric)
            except Exception:
                pass

        return valid_metrics

    def create_table_plot(self):
        """

        :return:
        """
        metrics = self.status.find_one({'type': 'MODEL_VALIDATION'})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(
                title_text='No data currently available for table mean metrics')
            table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        valid_metrics = self._scalar_metrics(metrics)
        if valid_metrics == []:
            fig = go.Figure(data=[])
            fig.update_layout(title_text='No scalar metrics found')
            table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        all_vals = []
        models = []
        for metric in valid_metrics:
            validations = {}
            for post in self.status.find({'type': 'MODEL_VALIDATION'}):
                e = json.loads(post['data'])
                try:
                    validations[e['modelId']].append(
                        float(json.loads(e['data'])[metric]))
                except KeyError:
                    validations[e['modelId']] = [
                        float(json.loads(e['data'])[metric])]

            vals = []
            models = []
            for model, data in validations.items():
                vals.append(numpy.mean(data))
                models.append(model)
            all_vals.append(vals)

        header_vals = valid_metrics
        models.reverse()
        values = [models]

        for vals in all_vals:
            vals.reverse()
            values.append(vals)

        fig = go.Figure(data=[go.Table(
            header=dict(values=['Model ID'] + header_vals,
                        line_color='darkslategray',
                        fill_color='lightskyblue',
                        align='left'),

            cells=dict(values=values,  # 2nd column
                       line_color='darkslategray',
                       fill_color='lightcyan',
                       align='left'))
        ])

        fig.update_layout(title_text='Summary: mean metrics')
        table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return table

    def create_timeline_plot(self):
        """

        :return:
        """
        trace_data = []
        x = []
        y = []
        base = []
        for p in self.status.find({'type': 'MODEL_UPDATE_REQUEST'}):
            e = json.loads(p['data'])
            cid = e['correlationId']
            for cc in self.status.find({'sender': p['sender'], 'type': 'MODEL_UPDATE'}):
                da = json.loads(cc['data'])
                if da['correlationId'] == cid:
                    cp = cc

            cd = json.loads(cp['data'])
            tr = datetime.strptime(e['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            tu = datetime.strptime(cd['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            ts = tu - tr
            base.append(tr.timestamp())
            x.append(ts.total_seconds() / 60.0)
            y.append(p['sender']['name'])

        trace_data.append(go.Bar(
            x=y,
            y=x,
            marker=dict(color='royalblue'),
            name="Training",
        ))

        x = []
        y = []
        base = []
        for p in self.status.find({'type': 'MODEL_VALIDATION_REQUEST'}):
            e = json.loads(p['data'])
            cid = e['correlationId']
            for cc in self.status.find({'sender': p['sender'], 'type': 'MODEL_VALIDATION'}):
                da = json.loads(cc['data'])
                if da['correlationId'] == cid:
                    cp = cc
            cd = json.loads(cp['data'])
            tr = datetime.strptime(e['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            tu = datetime.strptime(cd['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
            ts = tu - tr
            base.append(tr.timestamp())
            x.append(ts.total_seconds() / 60.0)
            y.append(p['sender']['name'])

        trace_data.append(go.Bar(
            x=y,
            y=x,
            marker=dict(color='lightskyblue'),
            name="Validation",
        ))

        layout = go.Layout(
            barmode='stack',
            showlegend=True,
        )

        fig = go.Figure(data=trace_data, layout=layout)
        fig.update_xaxes(title_text='Alliance/client')
        fig.update_yaxes(title_text='Time (Min)')
        fig.update_layout(title_text='Alliance timeline')
        timeline = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return timeline

    def create_client_training_distribution(self):
        """

        :return:
        """
        training = []
        for p in self.status.find({'type': 'MODEL_UPDATE'}):
            e = json.loads(p['data'])
            meta = json.loads(e['meta'])
            training.append(meta['exec_training'])

        if not training:
            return False
        fig = go.Figure(data=go.Histogram(x=training))
        fig.update_layout(
            title_text='Client model training time, mean: {}'.format(numpy.mean(training)))
        histogram = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return histogram

    def create_client_histogram_plot(self):
        """

        :return:
        """
        training = []
        for p in self.status.find({'type': 'MODEL_UPDATE'}):
            e = json.loads(p['data'])
            meta = json.loads(e['meta'])
            training.append(meta['exec_training'])

        fig = go.Figure()

        fig.update_layout(
            template="simple_white",
            xaxis=dict(title_text="Time (s)"),
            yaxis=dict(title_text='Number of updates'),
            title="Mean client training time: {}".format(
                numpy.mean(training)),
            # showlegend=True
        )
        if not training:
            return False

        fig.add_trace(go.Histogram(x=training))

        histogram_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return histogram_plot

    def create_client_plot(self):
        """

        :return:
        """
        processing = []
        upload = []
        download = []
        training = []
        for p in self.status.find({'type': 'MODEL_UPDATE'}):
            e = json.loads(p['data'])
            meta = json.loads(e['meta'])
            upload.append(meta['upload_model'])
            download.append(meta['fetch_model'])
            training.append(meta['exec_training'])
            processing.append(meta['processing_time'])

        fig = go.Figure()
        fig.update_layout(
            template="simple_white",
            title="Mean client processing time: {}".format(
                numpy.mean(processing)),
            showlegend=True
        )
        if not processing:
            return False
        data = [numpy.mean(training), numpy.mean(upload), numpy.mean(download)]
        labels = ["Training execution", "Model upload (to combiner)", "Model download (from combiner)"]
        fig.add_trace(go.Pie(labels=labels, values=data))

        client_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return client_plot

    def create_combiner_plot(self):
        """

        :return:
        """
        waiting = []
        aggregation = []
        model_load = []
        combination = []
        for round in self.mdb['control.round'].find():
            try:
                for combiner in round['combiners']:
                    data = combiner
                    stats = data['local_round']['1']
                    ml = stats['aggregation_time']['time_model_load']
                    ag = stats['aggregation_time']['time_model_aggregation']
                    combination.append(stats['time_combination'])
                    waiting.append(stats['time_combination'] - ml - ag)
                    model_load.append(ml)
                    aggregation.append(ag)
            except Exception:
                pass

        labels = ['Waiting for client updates',
                  'Aggregation', 'Loading model updates from disk']
        val = [numpy.mean(waiting), numpy.mean(
            aggregation), numpy.mean(model_load)]
        fig = go.Figure()

        fig.update_layout(
            template="simple_white",
            title="Mean combiner round time: {}".format(
                numpy.mean(combination)),
            showlegend=True
        )
        if not combination:
            return False
        fig.add_trace(go.Pie(labels=labels, values=val))
        combiner_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return combiner_plot

    def fetch_valid_metrics(self):
        """

        :return:
        """
        metrics = self.status.find_one({'type': 'MODEL_VALIDATION'})
        valid_metrics = self._scalar_metrics(metrics)
        return valid_metrics

    def create_box_plot(self, metric):
        """

        :param metric:
        :return:
        """
        metrics = self.status.find_one({'type': 'MODEL_VALIDATION'})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(title_text='No data currently available for metric distribution over  '
                                         'participants')
            box = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return box

        valid_metrics = self._scalar_metrics(metrics)
        if valid_metrics == []:
            fig = go.Figure(data=[])
            fig.update_layout(title_text='No scalar metrics found')
            box = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return box

        validations = {}
        for post in self.status.find({'type': 'MODEL_VALIDATION'}):
            e = json.loads(post['data'])
            try:
                validations[e['modelId']].append(
                    float(json.loads(e['data'])[metric]))
            except KeyError:
                validations[e['modelId']] = [
                    float(json.loads(e['data'])[metric])]

        # Make sure validations are plotted in chronological order
        model_trail = self.mdb.control.model.find_one({'key': 'model_trail'})
        model_trail_ids = model_trail['model']
        validations_sorted = []
        for model_id in model_trail_ids:
            try:
                validations_sorted.append(validations[model_id])
            except Exception:
                pass

        validations = validations_sorted

        box = go.Figure()

        y = []
        for j, acc in enumerate(validations):
            # x.append(j)
            y.append(numpy.mean([float(i) for i in acc]))
            if len(acc) >= 2:
                box.add_trace(go.Box(y=acc, name=str(j), marker_color="royalblue", showlegend=False,
                                     boxpoints=False))
            else:
                box.add_trace(go.Scatter(
                    x=[str(j)], y=[y[j]], showlegend=False))

        rounds = list(range(len(y)))
        box.add_trace(go.Scatter(
            x=rounds,
            y=y,
            name='Mean'
        ))

        box.update_xaxes(title_text='Rounds')
        box.update_yaxes(
            tickvals=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
        box.update_layout(title_text='Metric distribution over clients: {}'.format(metric),
                          margin=dict(l=20, r=20, t=45, b=20))
        box = json.dumps(box, cls=plotly.utils.PlotlyJSONEncoder)
        return box

    def create_round_plot(self):
        """

        :return:
        """
        trace_data = []
        metrics = self.round_time.find_one({'key': 'round_time'})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(
                title_text='No data currently available for round time')
            return False

        for post in self.round_time.find({'key': 'round_time'}):
            rounds = post['round']
            traces_data = post['round_time']

        trace_data.append(go.Scatter(
            x=rounds,
            y=traces_data,
            mode='lines+markers',
            name='Reducer'
        ))

        for rec in self.combiner_round_time.find({'key': 'combiner_round_time'}):
            c_traces_data = rec['round_time']

        trace_data.append(go.Scatter(
            x=rounds,
            y=c_traces_data,
            mode='lines+markers',
            name='Combiner'
        ))

        fig = go.Figure(data=trace_data)
        fig.update_xaxes(title_text='Round')
        fig.update_yaxes(title_text='Time (s)')
        fig.update_layout(title_text='Round time')
        round_t = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return round_t

    def create_cpu_plot(self):
        """

        :return:
        """
        metrics = self.psutil_usage.find_one({'key': 'cpu_mem_usage'})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(
                title_text='No data currently available for MEM and CPU usage')
            cpu = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        for post in self.psutil_usage.find({'key': 'cpu_mem_usage'}):
            cpu = post['cpu']
            mem = post['mem']
            ps_time = post['time']
            round = post['round']

        # Create figure with secondary y-axis
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig.add_trace(go.Scatter(
            x=ps_time,
            y=cpu,
            mode='lines+markers',
            name='CPU (%)'
        ))

        fig.add_trace(go.Scatter(
            x=ps_time,
            y=mem,
            mode='lines+markers',
            name='MEM (%)'
        ))

        fig.add_trace(go.Scatter(
            x=ps_time,
            y=round,
            mode='lines+markers',
            name='Round',
        ), secondary_y=True)

        fig.update_xaxes(title_text='Date Time')
        fig.update_yaxes(title_text='Percentage (%)')
        fig.update_yaxes(title_text="Round", secondary_y=True)
        fig.update_layout(title_text='CPU loads and memory usage')
        cpu = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return cpu



import os

from fedn.common.log_config import logger
from fedn.utils.process import run_process


class Dispatcher:
    """ Dispatcher class for compute packages.

    :param config: The configuration.
    :type config: dict
    :param dir: The directory to dispatch to.
    :type dir: str
    """

    def __init__(self, config, dir):
        """ Initialize the dispatcher."""
        self.config = config
        self.project_dir = dir

    def run_cmd(self, cmd_type):
        """ Run a command.

        :param cmd_type: The command type.
        :type cmd_type: str
        :return:
        """
        try:
            cmdsandargs = cmd_type.split(' ')

            cmd = [self.config['entry_points'][cmdsandargs[0]]['command']]

            # remove the first element,  that is not a file but a command
            args = cmdsandargs[1:]

            # shell (this could be a venv, TODO: parametrize)
            if os.name == "nt":
                shell = []
            else:
                shell = ['/bin/sh', '-c']

            # add the corresponding process defined in project.yaml and append arguments from invoked command
            args = shell + [' '.join(cmd + args)]
            run_process(args=args, cwd=self.project_dir)

            logger.info('Done executing {}'.format(cmd_type))
        except IndexError:
            message = "No such argument or configuration to run."
            logger.error(message)



from fedn.network.loadbalancer.loadbalancerbase import LoadBalancerBase


class LeastPacked(LoadBalancerBase):
    """ Load balancer that selects the first available combiner.

    :param network: A handle to the network.
    :type network: class: `fedn.network.api.network.Network`
    """

    def __init__(self, network):
        super().__init__(network)

    def find_combiner(self):
        """ Find the first available combiner. """

        for combiner in self.network.get_combiners():
            if combiner.allowing_clients():
                return combiner
        return None



from enum import Enum


class ReducerState(Enum):
    """ Enum for representing the state of a reducer."""
    setup = 1
    idle = 2
    instructing = 3
    monitoring = 4


def ReducerStateToString(state):
    """ Convert ReducerState to string.

    :param state: The state.
    :type state: :class:`fedn.network.state.ReducerState`
    :return: The state as string.
    :rtype: str
    """
    if state == ReducerState.setup:
        return "setup"
    if state == ReducerState.idle:
        return "idle"
    if state == ReducerState.instructing:
        return "instructing"
    if state == ReducerState.monitoring:
        return "monitoring"

    return "UNKNOWN"


def StringToReducerState(state):
    """ Convert string to ReducerState.

    :param state: The state as string.
    :type state: str
    :return: The state.
    :rtype: :class:`fedn.network.state.ReducerState`
    """
    if state == "setup":
        return ReducerState.setup
    if state == "idle":
        return ReducerState.idle
    elif state == "instructing":
        return ReducerState.instructing
    elif state == "monitoring":
        return ReducerState.monitoring



import io

from minio import Minio
from minio.error import InvalidResponseError
from urllib3.poolmanager import PoolManager

from fedn.common.log_config import logger
from fedn.network.storage.s3.base import RepositoryBase


class MINIORepository(RepositoryBase):
    """ Class implementing Repository for MinIO. """

    client = None

    def __init__(self, config):
        """ Initialize object.

        :param config: Dictionary containing configuration for credentials and bucket names.
        :type config: dict
        """

        super().__init__()
        self.name = "MINIORepository"

        if config['storage_secure_mode']:
            manager = PoolManager(
                num_pools=100, cert_reqs='CERT_NONE', assert_hostname=False)
            self.client = Minio("{0}:{1}".format(config['storage_hostname'], config['storage_port']),
                                access_key=config['storage_access_key'],
                                secret_key=config['storage_secret_key'],
                                secure=config['storage_secure_mode'], http_client=manager)
        else:
            self.client = Minio("{0}:{1}".format(config['storage_hostname'], config['storage_port']),
                                access_key=config['storage_access_key'],
                                secret_key=config['storage_secret_key'],
                                secure=config['storage_secure_mode'])
            logger.warning(
                "S3/MINIO RUNNING IN **INSECURE** MODE!")

    def set_artifact(self, instance_name, instance, bucket, is_file=False):

        if is_file:
            self.client.fput_object(bucket, instance_name, instance)
        else:
            try:
                self.client.put_object(
                    bucket, instance_name, io.BytesIO(instance), len(instance))
            except Exception as e:
                raise Exception("Could not load data into bytes {}".format(e))

        return True

    def get_artifact(self, instance_name, bucket):

        try:
            data = self.client.get_object(bucket, instance_name)
            return data.read()
        except Exception as e:
            raise Exception("Could not fetch data from bucket, {}".format(e))
        finally:
            data.close()
            data.release_conn()

    def get_artifact_stream(self, instance_name, bucket):

        try:
            data = self.client.get_object(bucket, instance_name)
            return data
        except Exception as e:
            raise Exception("Could not fetch data from bucket, {}".format(e))

    def list_artifacts(self, bucket):
        """ List all objects in bucket.

        :param bucket: Name of the bucket
        :type bucket: str
        :return: A list of object names
        """
        objects = []
        try:
            objs = self.client.list_objects(bucket)
            for obj in objs:
                objects.append(obj.object_name)
        except Exception:
            raise Exception(
                "Could not list models in bucket {}".format(bucket))
        return objects

    def delete_artifact(self, instance_name, bucket):
        """ Delete object with name instance_name from buckets.

        :param instance_name: The object name
        :param bucket: Buckets to delete from
        :type bucket: str
        """

        try:
            self.client.remove_object(bucket, instance_name)
        except InvalidResponseError as err:
            logger.error('Could not delete artifact: {0} err: {1}'.format(instance_name, err))
            pass

    def create_bucket(self, bucket_name):
        """ Create a new bucket. If bucket exists, do nothing.

        :param bucket_name: The name of the bucket
        :type bucket_name: str
        """
        found = self.client.bucket_exists(bucket_name)

        if not found:
            try:
                self.client.make_bucket(bucket_name)
            except InvalidResponseError:
                raise



from setuptools import find_packages, setup

setup(
    name='fedn',
    version='0.8.0',
    description="""Scaleout Federated Learning""",
    author='Scaleout Systems AB',
    author_email='contact@scaleoutsystems.com',
    url='https://www.scaleoutsystems.com',
    py_modules=['fedn'],
    python_requires='>=3.8,<3.11',
    install_requires=[
        "PyYAML>=5.4",
        "requests",
        "urllib3>=1.26.4",
        "minio",
        "python-slugify",
        "grpcio~=1.57.0",
        "grpcio-tools~=1.57.0",
        "numpy>=1.21.6",
        "protobuf",
        "pymongo",
        "Flask",
        "Flask-WTF",
        "pyjwt",
        "pyopenssl",
        "ttictoc",
        "psutil",
        "click==8.0.1",
        "jinja2",
        "plotly",
        "pandas",
        "bokeh<3.0.0",
        "networkx",
        "grpcio-health-checking~=1.57.0"
    ],
    license='Apache 2.0',
    zip_safe=False,
    entry_points={
        'console_scripts': ["fedn=cli:main"]
    },
    keywords='Federated learning',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



# This file contains the Connector class for assigning client to the FEDn network via the discovery service (REST-API).
# The Connector class is used by the Client class in fedn/network/clients/client.py.
# Once assigned, the client will retrieve combiner assignment from the discovery service.
# The discovery service will also add the client to the statestore.
#
#
import enum

import requests

from fedn.common.log_config import logger


class Status(enum.Enum):
    """ Enum for representing the status of a client assignment."""
    Unassigned = 0
    Assigned = 1
    TryAgain = 2
    UnAuthorized = 3
    UnMatchedConfig = 4


class ConnectorClient:
    """ Connector for assigning client to a combiner in the FEDn network.

    :param host: host of discovery service
    :type host: str
    :param port: port of discovery service
    :type port: int
    :param token: token for authentication
    :type token: str
    :param name: name of client
    :type name: str
    :param remote_package: True if remote package is used, False if local
    :type remote_package: bool
    :param force_ssl: True if https is used, False if http
    :type force_ssl: bool
    :param verify: True if certificate is verified, False if not
    :type verify: bool
    :param combiner: name of preferred combiner
    :type combiner: str
    :param id: id of client
    """

    def __init__(self, host, port, token, name, remote_package, force_ssl=False, verify=False, combiner=None, id=None):

        self.host = host
        self.port = port
        self.token = token
        self.name = name
        self.verify = verify
        self.preferred_combiner = combiner
        self.id = id
        self.package = 'remote' if remote_package else 'local'

        # for https we assume a an ingress handles permanent redirect (308)
        if force_ssl:
            self.prefix = "https://"
        else:
            self.prefix = "http://"
        if self.port:
            self.connect_string = "{}{}:{}".format(
                self.prefix, self.host, self.port)
        else:
            self.connect_string = "{}{}".format(
                self.prefix, self.host)

        logger.info("Setting connection string to {}.".format(self.connect_string))

    def assign(self):
        """
        Connect client to FEDn network discovery service, ask for combiner assignment.

        :return: Tuple with assingment status, combiner connection information if sucessful, else None.
        :rtype: tuple(:class:`fedn.network.clients.connect.Status`, str)
        """
        try:
            retval = None
            payload = {'client_id': self.name, 'preferred_combiner': self.preferred_combiner}

            retval = requests.post(self.connect_string + '/add_client',
                                   json=payload,
                                   verify=self.verify,
                                   allow_redirects=True,
                                   headers={'Authorization': 'Token {}'.format(self.token)})
        except Exception as e:
            print('***** {}'.format(e), flush=True)
            return Status.Unassigned, {}

        if retval.status_code == 400:
            # Get error messange from response
            reason = retval.json()['message']
            return Status.UnMatchedConfig, reason

        if retval.status_code == 401:
            reason = "Unauthorized connection to reducer, make sure the correct token is set"
            return Status.UnAuthorized, reason

        if retval.status_code >= 200 and retval.status_code < 204:
            if retval.json()['status'] == 'retry':
                if 'message' in retval.json():
                    reason = retval.json()['message']
                else:
                    reason = "Reducer was not ready. Try again later."

                return Status.TryAgain, reason

            reducer_package = retval.json()['package']
            if reducer_package != self.package:
                reason = "Unmatched config of compute package between client and reducer.\n" +\
                    "Reducer uses {} package and client uses {}.".format(
                        reducer_package, self.package)
                return Status.UnMatchedConfig, reason

            return Status.Assigned, retval.json()

        return Status.Unassigned, None



import glob
import os
from os.path import basename, dirname, isfile

from fedn.network.api.client import APIClient

# flake8: noqa


modules = glob.glob(dirname(__file__) + "/*.py")
__all__ = [basename(f)[:-3] for f in modules if isfile(f)
           and not f.endswith('__init__.py')]


_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_data(path):
    """

    :param path:
    :return:
    """
    return os.path.join(_ROOT, 'data', path)



import json

import requests

__all__ = ['APIClient']


class APIClient:
    """ An API client for interacting with the statestore and controller.

    :param host: The host of the api server.
    :type host: str
    :param port: The port of the api server.
    :type port: int
    :param secure: Whether to use https.
    :type secure: bool
    :param verify: Whether to verify the server certificate.
    :type verify: bool
    """

    def __init__(self, host, port, secure=False, verify=False):
        self.host = host
        self.port = port
        self.secure = secure
        self.verify = verify

    def _get_url(self, endpoint):
        if self.secure:
            protocol = 'https'
        else:
            protocol = 'http'
        return f'{protocol}://{self.host}:{self.port}/{endpoint}'

    def get_model_trail(self):
        """ Get the model trail.

        :return: The model trail as dict including commit timestamp.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_model_trail'), verify=self.verify)
        return response.json()

    def list_models(self, session_id=None):
        """ Get all models from the statestore.

        :return: All models.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_models'), params={'session_id': session_id}, verify=self.verify)
        return response.json()

    def list_clients(self):
        """ Get all clients from the statestore.

        return: All clients.
        rtype: dict
        """
        response = requests.get(self._get_url('list_clients'))
        return response.json()

    def get_active_clients(self, combiner_id):
        """ Get all active clients from the statestore.

        :param combiner_id: The combiner id to get active clients for.
        :type combiner_id: str
        :return: All active clients.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_active_clients'), params={'combiner': combiner_id}, verify=self.verify)
        return response.json()

    def get_client_config(self, checksum=True):
        """ Get the controller configuration. Optionally include the checksum.
        The config is used for clients to connect to the controller and ask for combiner assignment.

        :param checksum: Whether to include the checksum of the package.
        :type checksum: bool
        :return: The client configuration.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_client_config'), params={'checksum': checksum}, verify=self.verify)
        return response.json()

    def list_combiners(self):
        """ Get all combiners in the network.

        :return: All combiners with info.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_combiners'))
        return response.json()

    def get_combiner(self, combiner_id):
        """ Get a combiner from the statestore.

        :param combiner_id: The combiner id to get.
        :type combiner_id: str
        :return: The combiner info.
        :rtype: dict
        """
        response = requests.get(self._get_url(f'get_combiner?combiner={combiner_id}'), verify=self.verify)
        return response.json()

    def list_rounds(self):
        """ Get all rounds from the statestore.

        :return: All rounds with config and metrics.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_rounds'))
        return response.json()

    def get_round(self, round_id):
        """ Get a round from the statestore.

        :param round_id: The round id to get.
        :type round_id: str
        :return: The round config and metrics.
        :rtype: dict
        """
        response = requests.get(self._get_url(f'get_round?round_id={round_id}'), verify=self.verify)
        return response.json()

    def start_session(self, session_id=None, aggregator='fedavg', model_id=None, round_timeout=180, rounds=5, round_buffer_size=-1, delete_models=True,
                      validate=True, helper='numpyhelper', min_clients=1, requested_clients=8):
        """ Start a new session.

        :param session_id: The session id to start.
        :type session_id: str
        :param aggregator: The aggregator plugin to use.
        :type aggregator: str
        :param model_id: The id of the initial model.
        :type model_id: str
        :param round_timeout: The round timeout to use in seconds.
        :type round_timeout: int
        :param rounds: The number of rounds to perform.
        :type rounds: int
        :param round_buffer_size: The round buffer size to use.
        :type round_buffer_size: int
        :param delete_models: Whether to delete models after each round at combiner (save storage).
        :type delete_models: bool
        :param validate: Whether to validate the model after each round.
        :type validate: bool
        :param helper: The helper type to use.
        :type helper: str
        :param min_clients: The minimum number of clients required.
        :type min_clients: int
        :param requested_clients: The requested number of clients.
        :type requested_clients: int
        :return: A dict with success or failure message and session config.
        :rtype: dict
        """
        response = requests.post(self._get_url('start_session'), json={
            'session_id': session_id,
            'aggregator': aggregator,
            'model_id': model_id,
            'round_timeout': round_timeout,
            'rounds': rounds,
            'round_buffer_size': round_buffer_size,
            'delete_models': delete_models,
            'validate': validate,
            'helper': helper,
            'min_clients': min_clients,
            'requested_clients': requested_clients
        }, verify=self.verify
        )
        return response.json()

    def list_sessions(self):
        """ Get all sessions from the statestore.

        :return: All sessions in dict.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_sessions'), verify=self.verify)
        return response.json()

    def get_session(self, session_id):
        """ Get a session from the statestore.

        :param session_id: The session id to get.
        :type session_id: str
        :return: The session as a json object.
        :rtype: dict
        """
        response = requests.get(self._get_url(f'get_session?session_id={session_id}'), self.verify)
        return response.json()

    def session_is_finished(self, session_id):
        """ Check if a session with id session_id has finished.

        :param session_id: The session id to get.
        :type session_id: str
        :return: The session as a json object.
        :rtype: dict
        """
        try:
            status = self.get_session(session_id)['status']
            if status == 'Finished':
                return True
            else:
                return False
        except json.JSONDecodeError:
            # Could happen if the session has not been written to db yet
            return False
        except Exception:
            raise

    def set_package(self, path: str, helper: str, name: str = None, description: str = None):
        """ Set the compute package in the statestore.

        :param path: The file path of the compute package to set.
        :type path: str
        :param helper: The helper type to use.
        :type helper: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, 'rb') as file:
            response = requests.post(self._get_url('set_package'), files={'file': file}, data={
                                     'helper': helper, 'name': name, 'description': description}, verify=self.verify)
        return response.json()

    def get_package(self):
        """ Get the compute package from the statestore.

        :return: The compute package with info.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_package'), verify=self.verify)
        return response.json()

    def list_compute_packages(self):
        """ Get all compute packages from the statestore.

        :return: All compute packages with info.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_compute_packages'), verify=self.verify)
        return response.json()

    def download_package(self, path):
        """ Download the compute package.

        :param path: The path to download the compute package to.
        :type path: str
        :return: Message with success or failure.
        :rtype: dict
        """
        response = requests.get(self._get_url('download_package'), verify=self.verify)
        if response.status_code == 200:
            with open(path, 'wb') as file:
                file.write(response.content)
            return {'success': True, 'message': 'Package downloaded successfully.'}
        else:
            return {'success': False, 'message': 'Failed to download package.'}

    def get_package_checksum(self):
        """ Get the checksum of the compute package.

        :return: The checksum.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_package_checksum'), verify=self.verify)
        return response.json()

    def get_latest_model(self):
        """ Get the latest model from the statestore.

        :return: The latest model id.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_latest_model'), verify=self.verify)
        return response.json()

    def get_initial_model(self):
        """ Get the initial model from the statestore.

        :return: The initial model id.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_initial_model'), verify=self.verify)
        return response.json()

    def set_initial_model(self, path):
        """ Set the initial model in the statestore and upload to model repository.

        :param path: The file path of the initial model to set.
        :type path: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, 'rb') as file:
            response = requests.post(self._get_url('set_initial_model'), files={'file': file}, verify=self.verify)
        return response.json()

    def get_controller_status(self):
        """ Get the status of the controller.

        :return: The status of the controller.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_controller_status'), verify=self.verify)
        return response.json()

    def get_events(self, **kwargs):
        """ Get the events from the statestore. Pass kwargs to filter events.

        :return: The events in dict
        :rtype: dict
        """
        response = requests.get(self._get_url('get_events'), params=kwargs, verify=self.verify)
        return response.json()

    def list_validations(self, **kwargs):
        """ Get all validations from the statestore. Pass kwargs to filter validations.

        :return: All validations in dict.
        :rtype: dict
        """
        response = requests.get(self._get_url('list_validations'), params=kwargs, verify=self.verify)
        return response.json()



""" Module handling storage of objects in S3-compatible object storage. This functionality is used by the controller
to store global models in the model trail in persistent storage. Currently implemented for MinIO, but a '
developer can extend the framwork by implemeting the interface in base.py. """
